"""
Project : Simplified IDEA algorithm
Authors : Rayyan BELHI, Dario GENGA
Date : 2022
"""

NIBBLE_BIT_SIZE = 16
MODULO_FOR_ADD = 2**4  # = 16
MODULO_FOR_MUL = MODULO_FOR_ADD + 1  # = 2^4 + 1 = 17
NIBBLE_STRING_SIZE = 4
BINARY_BASE = 2
STRING_BINARY_FORMAT = '04b'
NB_SUBKEYS_BY_ROUND = 6
NB_SUBKEYS_FINAL_ROUND = 4


def modulo(a, b):
    """Compute a mod b.

    Args:
        a (int): The number a
        b (int): The number b

    Returns:
         int: The result of a mod b.
    """

    # mod(a, b) = a - b * floor(a / b)
    res = a - b * (a // b)
    return res


def add_mod(a, b):
    """Compute (a+b) mod 16.

    Args:
        a (str or int): The number a
        b (str or int): The number b

    Returns:
         str: A nibble string of (a+b) mod 16.
    """

    if type(a) == str:
        a = int(a, BINARY_BASE)

    if type(b) == str:
        b = int(b, BINARY_BASE)

    tmp = a + b
    res = modulo(tmp, MODULO_FOR_ADD)
    return format(res, STRING_BINARY_FORMAT)


def mul_mod(a, b):
    """Compute (a*b) mod 17.

    Args:
        a (str or int): The number a
        b (str or int): The number b

    Returns:
         str: A nibble string of (a*b) mod 17.
    """

    if type(a) == str:
        a = int(a, BINARY_BASE)

    if type(b) == str:
        b = int(b, BINARY_BASE)

    # If we have a nibble of 0000, then this corresponds to 16 (base 10) that is congruent to -1 modulo 17
    if a == 0:
        a = NIBBLE_BIT_SIZE
    if b == 0:
        b = NIBBLE_BIT_SIZE

    tmp = a * b
    res = modulo(tmp, MODULO_FOR_MUL)
    # Because we work with nibbles and  2^4 + 1 = 17, if we obtain 16 as result this will correspond to 10000 in binary.
    # This value can't be stored in a nibble, so we must exclude it
    if res == NIBBLE_BIT_SIZE:
        res = 0

    return format(res, STRING_BINARY_FORMAT)


def xor(a, b):
    """Compute a XOR b.

    Args:
        a (str or int): The number a
        b (str or int): The number b

    Returns:
         str: The result of a XOR b.
    """

    res = ''

    # Format to binary string
    if type(a) != str:
        a = format(a, STRING_BINARY_FORMAT)

    if type(b) != str:
        b = format(b, STRING_BINARY_FORMAT)

    # Adjust the size of a and b
    if len(a) != len(b):
        # Add leading zero to the smallest value
        if len(a) > len(b):
            tmp = a
            a = b
            b = tmp

        if len(a) < len(b):
            # Create a list with the first element containing the leading zeroes and the second the initial value
            value_as_list = ((len(b) - len(a)) * '0', a)
            # Then concatenate the two elements into a string
            a = ''.join(value_as_list)

    # Parse the strings and apply the XOR operation
    i = 0
    size = len(a)

    while i < size:
        if a[i] == b[i]:
            res += '0'
        else:
            res += '1'
        i += 1

    return res


def egcd(a, b):
    """Get the greatest common divisor of a and b by using the euclidean division

    Args:
        a (int): The number a
        b (int): The number b

    Returns:
         int: The gcd of a and b.
    """

    if a < b:
        tmp = a
        a = b
        b = tmp

    r = b
    old_r = r

    while r != 0:
        old_r = r
        r = modulo(a, b)
        a = b
        b = r

    return old_r


def get_bezout_coefficients(a, b):
    """Find the Bézout coefficients of a and b by using the extended euclidean algorithm.

    Args:
        a (int): The number a
        b (int): The number b.

    Returns:
        tuple: Bézout coefficients.
    """

    i = 2
    q = [0, 0]
    r = [a, b]
    x = [1, 0]
    y = [0, 1]

    while True:
        r.append(modulo(r[i - 2], r[i - 1]))

        # Continue until the rest is equal to 0
        if r[i] == 0:
            break

        q.append(r[i - 2] // r[i - 1])
        x.append(x[i - 2] - q[i] * x[i - 1])
        y.append(y[i - 2] - q[i] * y[i - 1])

        i += 1

    return x[-1], y[-1]


def modular_inverse(a, b):
    """Compute the modular inverse of a number a modulo b.

    Args:
        a (str or int): The number to reverse.
        b (str or int): The modulo.

    Returns:
        str: A nibble string of the reversed number.
    """
    if type(a) == str:
        a = int(a, BINARY_BASE)

    x, y = get_bezout_coefficients(a, b)
    bezout = a * x + b * y

    if bezout != egcd(a, b):
        print("Error detected in the Bézout's identity")
        return None

    if bezout == 1 and modulo(a * x, b) == 1:
        return format(modulo(x, b), STRING_BINARY_FORMAT)
    return None


def inverse_add_mod(a):
    """Get the inverse of nibble for addition modulo 16

    Args:
        a (str or int): The number to reverse.

    Returns:
        str: A nibble string of the inverse.
    """
    if type(a) == str:
        a = int(a, BINARY_BASE)

    res = modulo(-a, NIBBLE_BIT_SIZE)

    return format(res, STRING_BINARY_FORMAT)


def shift_key(key, x=6):
    """Shift the x bits starting from the left of the key to the end

    Args:
        key (str): The key used for the algorithm.
        x (int): The size of the shift.

    Returns:
        str: The key shifted.
    """
    bits_to_shift = key[0:x]
    new_key = key[x:] + bits_to_shift
    return new_key


def create_subkeys_table(key):
    """Create the table of the 28 subkeys used for the IDEA encryption

    Args:
        key (str): The key used for the algorithm.

    Returns:
        2D array: A 2d array of the subkeys.
    """
    subkeys = []
    subkeys_round = []
    nibble = ''

    current_round = 1
    total_rounds = 4

    while current_round <= total_rounds:
        # Decompose the key into 8 nibbles
        for bit in key:
            nibble += bit
            if len(nibble) >= NIBBLE_STRING_SIZE:
                subkeys_round.append(nibble)
                nibble = ''

                if len(subkeys_round) == NB_SUBKEYS_BY_ROUND:
                    subkeys.append(subkeys_round)
                    subkeys_round = []

                # The final round contains less subkeys
                if current_round == total_rounds and len(subkeys_round) == NB_SUBKEYS_FINAL_ROUND:
                    subkeys.append(subkeys_round)
                    # We must break the loop when we have all subkeys
                    break

        # Shift the key for the next round
        key = shift_key(key)
        current_round += 1

    return subkeys


def create_decryption_subkeys_table(subkeys_table):
    """Create the table of the 28 subkeys used for the IDEA decryption

    Args:
        subkeys_table ([][]): A 2d array of the subkeys.

    Returns:
        2D array: A 2d array of the decryption subkeys.
    """
    decryption_subkeys = []

    remaining_round = 4
    # The first decryption subkeys is created by the first encryption key of the last round
    # So we will loop in reversed
    while remaining_round >= 0:
        subkeys_round = []
        k1 = modular_inverse(subkeys_table[remaining_round][0], MODULO_FOR_MUL)
        k2 = inverse_add_mod(subkeys_table[remaining_round][1])
        k3 = inverse_add_mod(subkeys_table[remaining_round][2])
        k4 = modular_inverse(subkeys_table[remaining_round][3], MODULO_FOR_MUL)

        subkeys_round.append(k1)
        subkeys_round.append(k2)
        subkeys_round.append(k3)
        subkeys_round.append(k4)

        if remaining_round > 0:
            # The last two decryption subkeys of a round correspond to the last
            # two encryption subkeys of the previous round
            k5 = subkeys_table[remaining_round - 1][4]
            k6 = subkeys_table[remaining_round - 1][5]
            subkeys_round.append(k5)
            subkeys_round.append(k6)

        decryption_subkeys.append(subkeys_round)
        remaining_round -= 1

    return decryption_subkeys


def encrypt(plaintext, subkeys):
    """Encrypt the text with the subkeys by using the simplified IDEA algorithm

    Args:
        plaintext (str): The text to encrypt
        subkeys ([][]): A 2d array of the subkeys.

    Returns:
        str: The encrypted text..
    """
    input_block = plaintext

    current_round = 0
    total_rounds = 4

    while current_round <= total_rounds:
        # 1. Multiply X1 and the first subkey Z1
        res_1 = mul_mod(input_block[0:4], subkeys[current_round][0])
        # 2. Add X2 and the second subkey Z2
        res_2 = add_mod(input_block[4:8], subkeys[current_round][1])
        # 3. Add X3 and the third subkey Z3
        res_3 = add_mod(input_block[8:12], subkeys[current_round][2])
        # 4. Multiply X4 and the fourth subkey Z4
        res_4 = mul_mod(input_block[12:16], subkeys[current_round][3])

        # Don't do the next steps if this is the final round
        if current_round < total_rounds:
            # 5. Bitwise XOR the results of steps 1 and 3
            res_5 = xor(res_1, res_3)
            # 6. Bitwise XOR the results of steps 2 and 4
            res_6 = xor(res_2, res_4)
            # 7. Multiply the result of step 5 and the fifth subkey Z5
            res_7 = mul_mod(res_5, subkeys[current_round][4])
            # 8. Add the results of step 6 and 7
            res_8 = add_mod(res_6, res_7)
            # 9. Multiply the result of step 8 and the sixth subkey Z6
            res_9 = mul_mod(res_8, subkeys[current_round][5])
            # 10. Add the results of steps 7 and 9
            res_10 = add_mod(res_7, res_9)
            # 11. Bitwise XOR the results of steps 1 and 9
            res_11 = xor(res_1, res_9)
            # 12. Bitwise XOR the results of steps 2 and 10
            res_12 = xor(res_2, res_10)
            # 13. Bitwise XOR the results of steps 3 and 9
            res_13 = xor(res_3, res_9)
            # 14. Bitwise XOR the results of steps 4 and 10
            res_14 = xor(res_4, res_10)
            # Create the input for the next round with the final results
            input_block = res_11 + res_12 + res_13 + res_14

        current_round += 1

    ciphertext = res_1 + res_2 + res_3 + res_4
    return ciphertext


def decrypt(ciphertext, subkeys_table):
    """Decrypt the ciphertext with the subkeys table by using the simplified IDEA algorithm

    Args:
        ciphertext (str): The text to decrypt.
        subkeys_table ([][]): A 2d array of the subkeys.

    Returns:
        str: The decrypted text.
    """

    # To decrypt the ciphertext, we just have to encrypt it by using the decryption subkeys table
    decryption_keys = create_decryption_subkeys_table(subkeys_table)
    return encrypt(ciphertext, decryption_keys)


def main():
    plaintext = '1001110010101100'
    key = '11011100011011110011111101011001'
    expected_ciphertext = '1011101101001011'

    print("The original message is : " + plaintext)
    print("The key is : " + key)

    subkeys = create_subkeys_table(key)

    print("Starting the encryption...")

    ciphertext = encrypt(plaintext, subkeys)

    print("The ciphertext obtained is : " + ciphertext)

    if ciphertext == expected_ciphertext:
        print("The encryption has been successful !")
    else:
        print("An error has occurred during the encryption...")

    print("Starting the decryption...")

    decrypted_text = decrypt(ciphertext, subkeys)

    print("The decrypted message is : " + decrypted_text)

    if decrypted_text == plaintext:
        print("The decryption has been successful !")
    else:
        print("An error has occurred during the decryption...")


if __name__ == "__main__":
    main()

